@@ -26,8 +26,9 @@
 #include <tbb/parallel_for.h>
 #endif
 
+
 //#include "R.h"
-#include "Rcpp.h" // TODO Remove
+//#include "Rcpp.h" // TODO Remove
 
 #ifdef CYCLOPS_DEBUG_TIMING
 	#include "Timing.h"
@@ -85,6 +86,7 @@ ModelSpecifics<BaseModel,RealType>::ModelSpecifics(const ModelData<RealType>& in
 	{
     	// Do nothing
 	// TODO Memory allocation here
+	// std::cerr << "ctor ModelSpecifics \n";
 
 #ifdef CYCLOPS_DEBUG_TIMING
 	auto now = bsccs::chrono::system_clock::now();
@@ -96,13 +98,17 @@ ModelSpecifics<BaseModel,RealType>::ModelSpecifics(const ModelData<RealType>& in
 }
 
 template <class BaseModel, typename RealType>
-AbstractModelSpecifics* ModelSpecifics<BaseModel,RealType>::clone() const {
+AbstractModelSpecifics* ModelSpecifics<BaseModel,RealType>::clone(ComputeDeviceArguments computeDevice) const {
 	return new ModelSpecifics<BaseModel,RealType>(modelData);
 }
 
 template <class BaseModel, typename RealType>
 double ModelSpecifics<BaseModel,RealType>::getGradientObjective(bool useCrossValidation) {
 
+#ifdef CYCLOPS_DEBUG_TIMING
+    auto start = bsccs::chrono::steady_clock::now();
+#endif
+
 		auto& xBeta = getXBeta();
 
 		RealType criterion = 0;
@@ -115,7 +121,11 @@ double ModelSpecifics<BaseModel,RealType>::getGradientObjective(bool useCrossVal
 				criterion += xBeta[i] * hY[i];
             }
 		}
-
+#ifdef CYCLOPS_DEBUG_TIMING
+    auto end = bsccs::chrono::steady_clock::now();
+    ///////////////////////////"
+    duration["getGradObj       "] += bsccs::chrono::duration_cast<chrono::TimingUnits>(end - start).count();
+#endif
 		return static_cast<double> (criterion);
 	}
 
@@ -136,6 +146,7 @@ void ModelSpecifics<BaseModel,RealType>::printTiming() {
 template <class BaseModel,typename RealType>
 ModelSpecifics<BaseModel,RealType>::~ModelSpecifics() {
 	// TODO Memory release here
+	// std::cerr << "dtor ModelSpecifics \n";
 
 #ifdef CYCLOPS_DEBUG_TIMING
 
@@ -458,22 +469,28 @@ void ModelSpecifics<BaseModel,RealType>::setWeights(double* inWeights, double *c
     if (hYWeight.size() != K) {
         hYWeight.resize(K);
     }
-
+	if (hYWeightDouble.size() != K) {
+		hYWeightDouble.resize(K);
+	}
     if (BaseModel::isTwoWayScan) {
         hNtoK.resize(N + 1);
         int n = 0;
         for (size_t k = 0; k < K;) {
+            while (hKWeight[k] == static_cast<RealType>(0)) {
+                ++k;
+            }
             hNtoK[n] = k;
             int currentPid = hPid[k];
             do {
                 ++k;
-            } while (k < K && currentPid == hPid[k]);
+            } while (k < K && (currentPid == hPid[k] || hKWeight[k] == static_cast<RealType>(0)));
             ++n;
         }
         hNtoK[n] = K;
 
         for (size_t k = 0; k < K; ++k) {
             hYWeight[k] = cenWeights[k];
+            hYWeightDouble[k] = cenWeights[k];
         }
     }
 
@@ -588,6 +605,9 @@ void ModelSpecifics<BaseModel,RealType>::computeFixedTermsInLogLikelihood(bool u
 
 template <class BaseModel,typename RealType>
 void ModelSpecifics<BaseModel,RealType>::computeFixedTermsInGradientAndHessian(bool useCrossValidation) {
+#ifdef CYCLOPS_DEBUG_TIMING
+    auto start = bsccs::chrono::steady_clock::now();
+#endif
 	if (sortPid()) {
 		doSortPid(useCrossValidation);
 	}
@@ -600,6 +620,12 @@ void ModelSpecifics<BaseModel,RealType>::computeFixedTermsInGradientAndHessian(b
 	if (allocateNtoKIndices()) {
 		computeNtoKIndices(useCrossValidation);
 	}
+#ifdef CYCLOPS_DEBUG_TIMING
+    auto end = bsccs::chrono::steady_clock::now();
+    ///////////////////////////"
+    auto name = "compFixedGH      ";
+    duration[name] += bsccs::chrono::duration_cast<chrono::TimingUnits>(end - start).count();
+#endif
 }
 
 template <class BaseModel,typename RealType>
@@ -625,7 +651,7 @@ double ModelSpecifics<BaseModel,RealType>::getLogLikelihood(bool useCrossValidat
 			for (size_t i = 0; i < N; i++) {
 				// Weights modified in computeNEvents()
 				logLikelihood -= BaseModel::logLikeDenominatorContrib(hNWeight[i], accDenomPid[i]);
-            }
+			}
 		} else {  // TODO Unnecessary code duplication
 			for (size_t i = 0; i < N; i++) {
 				// Weights modified in computeNEvents()
@@ -652,10 +678,10 @@ double ModelSpecifics<BaseModel,RealType>::getLogLikelihood(bool useCrossValidat
 template <class BaseModel,typename RealType>
 double ModelSpecifics<BaseModel,RealType>::getPredictiveLogLikelihood(double* weights) {
 
-    std::vector<double> saveKWeight;
+	std::vector<double> saveKWeight;
 	if (BaseModel::cumulativeGradientAndHessian)	{
 
- 		// saveKWeight = hKWeight; // make copy
+ 	    // saveKWeight = hKWeight; // make copy
 	    if (saveKWeight.size() != K) {
 	        saveKWeight.resize(K);
 	    }
@@ -664,7 +690,7 @@ double ModelSpecifics<BaseModel,RealType>::getPredictiveLogLikelihood(double* we
 	    }
 
 		setPidForAccumulation(weights);
-		setWeights(weights, nullptr, true); // set new weights // TODO Possible error for gfr
+		setWeights(weights, BaseModel::isTwoWayScan ? hYWeightDouble.data() : nullptr, true); // set new weights // TODO Possible error for gfr
 		computeRemainingStatistics(true); // compute accDenomPid
 	}
 
@@ -682,7 +708,7 @@ double ModelSpecifics<BaseModel,RealType>::getPredictiveLogLikelihood(double* we
 
 	if (BaseModel::cumulativeGradientAndHessian) {
 		setPidForAccumulation(&saveKWeight[0]);
-	    setWeights(saveKWeight.data(), nullptr, true); // set old weights // TODO Possible eror for gfr
+		setWeights(saveKWeight.data(), BaseModel::isTwoWayScan ? hYWeightDouble.data() : nullptr, true); // set old weights // TODO Possible error for gfr
 		computeRemainingStatistics(true);
 	}
 
@@ -763,6 +789,9 @@ void ModelSpecifics<BaseModel,RealType>::computeGradientAndHessian(int index, do
 #endif
 #endif
 
+#ifdef CYCLOPS_DEBUG_TIMING_GPU_COX
+        duration["CPU GH           "] += bsccs::chrono::duration_cast<chrono::TimingUnits>(end - start).count();
+#endif
 }
 
 template <class RealType>
@@ -1496,13 +1525,20 @@ void ModelSpecifics<BaseModel,RealType>::computeNumeratorForGradient(int index,
 #endif
 #endif
 
-	if (BaseModel::cumulativeGradientAndHessian) { // cox
+	if (BaseModel::hasNtoKIndices || BaseModel::cumulativeGradientAndHessian) {
+
 		switch (hX.getFormatType(index)) {
 		case INDICATOR : {
 				IndicatorIterator<RealType> itI(*(sparseIndices)[index]);
 				for (; itI; ++itI) { // Only affected entries
 					numerPid[itI.index()] = static_cast<RealType>(0.0);
 				}
+/*
+				zeroVector(numerPid.data(), N);
+				if (BaseModel::efron) {
+					zeroVector(numerPid3.data(), N);
+				}
+*/
 				if (useWeights) {
 				    incrementNumeratorForGradientImpl<IndicatorIterator<RealType>, WeightedOperation>(index);
 				} else {
@@ -1518,6 +1554,16 @@ void ModelSpecifics<BaseModel,RealType>::computeNumeratorForGradient(int index,
 						numerPid2[itS.index()] = static_cast<RealType>(0.0); // TODO Does this invalid the cache line too much?
 					}
 				}
+/*
+				zeroVector(numerPid.data(), N);
+				if (BaseModel::hasTwoNumeratorTerms) { // Compile-time switch
+					zeroVector(numerPid2.data(), N);
+				}
+				if (BaseModel::efron) {
+					zeroVector(numerPid3.data(), N);
+					zeroVector(numerPid4.data(), N);
+				}
+*/
 				if (useWeights) {
 				    incrementNumeratorForGradientImpl<SparseIterator<RealType>, WeightedOperation>(index);
 				} else {
@@ -1530,6 +1576,12 @@ void ModelSpecifics<BaseModel,RealType>::computeNumeratorForGradient(int index,
 				if (BaseModel::hasTwoNumeratorTerms) { // Compile-time switch
 					zeroVector(numerPid2.data(), N);
 				}
+/*
+				if (BaseModel::efron) {
+					zeroVector(numerPid3.data(), N);
+					zeroVector(numerPid4.data(), N);
+				}
+*/
 				if (useWeights) {
 				    incrementNumeratorForGradientImpl<DenseIterator<RealType>, WeightedOperation>(index);
 				} else {
@@ -1542,6 +1594,12 @@ void ModelSpecifics<BaseModel,RealType>::computeNumeratorForGradient(int index,
 				if (BaseModel::hasTwoNumeratorTerms) { // Compile-time switch
 					zeroVector(numerPid2.data(), N);
 				}
+/*
+				if (BaseModel::efron) {
+					zeroVector(numerPid3.data(), N);
+					zeroVector(numerPid4.data(), N);
+				}
+*/
 				if (useWeights) {
 				    incrementNumeratorForGradientImpl<InterceptIterator<RealType>, WeightedOperation>(index);
 				} else {
@@ -1561,6 +1619,9 @@ void ModelSpecifics<BaseModel,RealType>::computeNumeratorForGradient(int index,
 #endif
 #endif
 
+#ifdef CYCLOPS_DEBUG_TIMING_GPU_COX
+	duration["CPU GH           "] += bsccs::chrono::duration_cast<chrono::TimingUnits>(end - start).count();
+#endif
 }
 
 template <class BaseModel,typename RealType> template <class IteratorType, class Weights>
@@ -1573,22 +1634,37 @@ void ModelSpecifics<BaseModel,RealType>::incrementNumeratorForGradientImpl(int i
 #endif
 
 	IteratorType it(hX, index);
-    for (; it; ++it) {
-        const int k = it.index();
-        incrementByGroup(numerPid.data(), hPid, k,
-                         Weights::isWeighted ?
-                             hKWeight[k] * BaseModel::gradientNumeratorContrib(it.value(), offsExpXBeta[k], hXBeta[k], hY[k]) :
-                             BaseModel::gradientNumeratorContrib(it.value(), offsExpXBeta[k], hXBeta[k], hY[k])
-        );
-        if (!IteratorType::isIndicator && BaseModel::hasTwoNumeratorTerms) {
-            incrementByGroup(numerPid2.data(), hPid, k,
-                             Weights::isWeighted ?
-                                 hKWeight[k] * BaseModel::gradientNumerator2Contrib(it.value(), offsExpXBeta[k]) :
-                                 BaseModel::gradientNumerator2Contrib(it.value(), offsExpXBeta[k])
-            );
-        }
-    }
-
+	for (; it; ++it) {
+		const int k = it.index();
+		incrementByGroup(numerPid.data(), hPid, k,
+                   Weights::isWeighted ?
+                       hKWeight[k] * BaseModel::gradientNumeratorContrib(it.value(), offsExpXBeta[k], hXBeta[k], hY[k]) :
+		               BaseModel::gradientNumeratorContrib(it.value(), offsExpXBeta[k], hXBeta[k], hY[k])
+		    );
+		if (!IteratorType::isIndicator && BaseModel::hasTwoNumeratorTerms) {
+			incrementByGroup(numerPid2.data(), hPid, k,
+                    Weights::isWeighted ?
+                        hKWeight[k] * BaseModel::gradientNumerator2Contrib(it.value(), offsExpXBeta[k]) :
+                        BaseModel::gradientNumerator2Contrib(it.value(), offsExpXBeta[k])
+                );
+		}
+/*
+		if (BaseModel::efron) {
+			incrementByGroup(numerPid3.data(), hPid, k,
+					Weights::isWeighted ?
+						hKWeight[k] * hY[k] * BaseModel::gradientNumeratorContrib(it.value(), offsExpXBeta[k], hXBeta[k], hY[k]) :
+						hY[k] * BaseModel::gradientNumeratorContrib(it.value(), offsExpXBeta[k], hXBeta[k], hY[k])
+				);
+		}
+		if (!IteratorType::isIndicator && BaseModel::efron) {
+			incrementByGroup(numerPid4.data(), hPid, k,
+                    Weights::isWeighted ?
+                        hKWeight[k] * hY[k] * BaseModel::gradientNumerator2Contrib(it.value(), offsExpXBeta[k]) :
+                        hY[k] * BaseModel::gradientNumerator2Contrib(it.value(), offsExpXBeta[k])
+                );
+		}
+*/
+	}
 #ifdef CYCLOPS_DEBUG_TIMING
 #ifdef CYCLOPS_DEBUG_TIMING_LOW
 	auto end = bsccs::chrono::steady_clock::now();
@@ -1656,6 +1732,10 @@ void ModelSpecifics<BaseModel,RealType>::updateXBeta(double delta, int index, bo
 #endif
 #endif
 
+#ifdef CYCLOPS_DEBUG_TIMING_GPU_COX
+	duration["CPU GH           "] += bsccs::chrono::duration_cast<chrono::TimingUnits>(end - start).count();
+#endif
+
 }
 
 template <class BaseModel,typename RealType> template <class IteratorType, class Weights>
@@ -1692,6 +1772,11 @@ inline void ModelSpecifics<BaseModel,RealType>::updateXBetaImpl(RealType realDel
                 RealType oldEntry = offsExpXBeta[k];
                 RealType newEntry = offsExpXBeta[k] = BaseModel::getOffsExpXBeta(hOffs.data(), hXBeta[k], hY[k], k);
                 incrementByGroup(denomPid.data(), hPid, k, (newEntry - oldEntry));
+/*
+                if (BaseModel::efron) {
+                    incrementByGroup(denomPid2.data(), hPid, k, hY[k]*(newEntry - oldEntry)); // Update denominators
+                }
+*/
             }
         }
     }
@@ -1716,7 +1801,11 @@ void ModelSpecifics<BaseModel,RealType>::computeRemainingStatisticsImpl() {
 
     if (BaseModel::likelihoodHasDenominator) {
         fillVector(denomPid.data(), N, BaseModel::getDenomNullValue());
-
+/*
+        if (BaseModel::efron) {
+        	fillVector(denomPid2.data(), N, BaseModel::getDenomNullValue());
+        }
+*/
         if (BaseModel::cumulativeGradientAndHessian) { // cox
             for (size_t k = 0; k < K; ++k) {
                 offsExpXBeta[k] = BaseModel::getOffsExpXBeta(hOffs.data(), xBeta[k], hY[k], k);
@@ -1729,6 +1818,11 @@ void ModelSpecifics<BaseModel,RealType>::computeRemainingStatisticsImpl() {
             for (size_t k = 0; k < K; ++k) {
                 offsExpXBeta[k] = BaseModel::getOffsExpXBeta(hOffs.data(), xBeta[k], hY[k], k);
                 incrementByGroup(denomPid.data(), hPid, k, offsExpXBeta[k]);
+/*
+                if (BaseModel::efron) {
+                    incrementByGroup(denomPid2.data(), hPid, k, hY[k]*offsExpXBeta[k]); // Update denominators
+                }
+*/
             }
         }
         computeAccumlatedDenominator(Weights::isWeighted);
@@ -1797,6 +1891,7 @@ void ModelSpecifics<BaseModel,RealType>::computeAccumlatedDenominator(bool useWe
     if (BaseModel::likelihoodHasDenominator && //The two switches should ideally be separated
         BaseModel::cumulativeGradientAndHessian) { // Compile-time switch
 
+
         if (accDenomPid.size() != (N + 1)) {
             accDenomPid.resize(N + 1, static_cast<RealType>(0));
         }
@@ -1836,6 +1931,7 @@ void ModelSpecifics<BaseModel,RealType>::computeAccumlatedDenominator(bool useWe
             }
         } // End two-way scan
     }
+
 }
 
 // ESK: This is no longer needed. Incorporated in incrementGradientAndHessian
@@ -1929,9 +2025,19 @@ void ModelSpecifics<BaseModel,RealType>::doSortPid(bool useCrossValidation) {
  */
 }
 
+template <class BaseModel,typename RealType>
+void ModelSpecifics<BaseModel,RealType>::getOriginalPid() {
+
+    hPidInternal =  hPidOriginal; // Make copy
+
+}
+template <class BaseModel,typename RealType>
+void ModelSpecifics<BaseModel,RealType>::setPidForAccumulation(const double* weights) {
+	setPidForAccumulationImpl(weights);
+}
 
 template <class BaseModel,typename RealType> template <typename AnyRealType>
-void ModelSpecifics<BaseModel,RealType>::setPidForAccumulation(const AnyRealType* weights) {
+void ModelSpecifics<BaseModel,RealType>::setPidForAccumulationImpl(const AnyRealType* weights) {
 
     hPidInternal =  hPidOriginal; // Make copy
     hPid = hPidInternal.data(); // Point to copy
@@ -2051,7 +2157,9 @@ void ModelSpecifics<BaseModel,RealType>::initialize(
     if (allocateXjX()) {
         hXjX.resize(J);
     }
-
+#ifdef CYCLOPS_DEBUG_TIMING_GPU_COX
+	auto start = std::chrono::steady_clock::now();
+#endif
     if (initializeAccumulationVectors()) {
         setPidForAccumulation(static_cast<double*>(nullptr)); // calls setupSparseIndices() before returning
     } else {
@@ -2059,7 +2167,11 @@ void ModelSpecifics<BaseModel,RealType>::initialize(
         // If true, then fill with pointers to CompressedDataColumn and do not delete in destructor
         setupSparseIndices(N); // Need to be recomputed when hPid change!
     }
-
+#ifdef CYCLOPS_DEBUG_TIMING_GPU_COX
+	auto end = std::chrono::steady_clock::now();
+	double timerPid = std::chrono::duration<double>(end - start).count();
+	std::cout << " OVERHEAD CCD setPid:  " << timerPid << " s \n";
+#endif
 
 
     size_t alignedLength = getAlignedLength(N + 1);
@@ -2068,8 +2180,11 @@ void ModelSpecifics<BaseModel,RealType>::initialize(
     // 	denomPid = numerPid + alignedLength; // Nested in denomPid allocation
     // 	numerPid2 = numerPid + 2 * alignedLength;
     denomPid.resize(alignedLength);
+    denomPid2.resize(alignedLength);
     numerPid.resize(alignedLength);
     numerPid2.resize(alignedLength);
+    numerPid3.resize(alignedLength);
+    numerPid4.resize(alignedLength);
 
     deviceInitialization();
 
