@@ -139,6 +139,8 @@ public:
 		// Do nothing
 	}
 
+	virtual PriorType getPriorType() {return PriorType::NONE;};
+
 	virtual const std::string getDescription() const = 0; // pure virtual
 
 	virtual double logDensity(const DoubleVector& beta, const int index,
@@ -179,6 +181,10 @@ public:
 		return "None";
 	}
 
+	PriorType getPriorType() {
+		return PriorType::NONE;
+	}
+
     double logDensity(const DoubleVector& beta, const int index, CyclicCoordinateDescent& ccd) const {
         return 0.0;
     }
@@ -262,6 +268,11 @@ public:
 		info << "Laplace(" << lambda << ")";
 		return info.str();
 	}
+
+	PriorType getPriorType() {
+		return PriorType::LAPLACE;
+	}
+
     double logDensity(const DoubleVector& beta, const int index, CyclicCoordinateDescent& ccd) const {
         auto x = beta[index];
         auto lambda = getLambda();
@@ -433,6 +444,10 @@ public:
 		return info.str();
 	}
 
+	PriorType getPriorType() {
+		return PriorType::NORMAL;
+	}
+
 	bool getIsRegularized() const {
 	    return true;
 	}
